@@ -20,28 +20,31 @@
 #define _LARGEFILE64_SOURCE
 #define _GNU_SOURCE
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdbool.h>
-#include <stdint.h>
-#include <unistd.h>
-#include <string.h>
-#include <fcntl.h>
-#include <limits.h>
+#include <ctype.h>
 #include <dirent.h>
-#include <mntent.h>
-#include <getopt.h>
 #include <err.h>
 #include <errno.h>
+#include <fcntl.h>
+#include <getopt.h>
+#include <limits.h>
+#include <mntent.h>
 #include <signal.h>
+#include <stdbool.h>
+#include <stdint.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <time.h>
+#include <unistd.h>
+
+#include <sys/fanotify.h>
+/* for MIN() */
+#include <sys/param.h>
 #include <sys/stat.h>
 #include <sys/statfs.h>
-#include <sys/fanotify.h>
-#include <sys/time.h>
 #include <sys/sysmacros.h>
+#include <sys/time.h>
 #include <sys/types.h>
-#include <ctype.h>
 
 #define BUFSIZE 256*1024
 
@@ -61,12 +64,15 @@
 static char* option_output = NULL;
 static long option_filter_mask = 0xffffffff;
 static long option_timeout = -1;
-static int option_current_mount = 0;
+static bool option_current_mount = false;
 static int option_timestamp = 0;
-static int option_user = 0;
+static bool option_user = false;
 static pid_t ignored_pids[1024];
 static unsigned int ignored_pids_len = 0;
 static char* option_comm = NULL;
+static bool option_json = false;
+static bool option_parents = false;
+static bool option_exe = false;
 
 /* --time alarm sets this to 0 */
 static volatile int running = 1;
@@ -225,6 +231,125 @@ show_pid (pid_t pid)
     return true;
 }
 
+/* if str is a valid UTF-8 string without need of any JSON escaping, return the
+   byte length, otherwise -1. */
+static inline int
+nonfunny_utf8_len (const char* str) {
+    const unsigned char* s = (unsigned char*)str;
+    int i = 0;
+    while (str[i] != 0) {
+        unsigned char c = s[i];
+        // Unescaped ASCII
+        if (0x20 <= c && c != '"' && c != '\\' && c <= 0x7e) {
+            i++; continue;
+        }
+        // it's ok to read s[i+1] since we know s[i] != 0
+        uint32_t mbc = c<<8 | s[i+1];
+        if (// 2-char: 110xxxxx 10xxxxxx
+            (mbc & 0xe0c0) == 0xc080 &&
+            // but not 1100000x 10xxxxxx (overlong)
+            (mbc & 0xfec0) != 0xc080) {
+            i+=2; continue;
+        }
+        if (s[i+1] == 0)
+            return -1;
+        // it's ok to read s[i+2] since we know s[i+1] != 0
+        mbc = mbc<<8 | s[i+2];
+        if (// 3-char: 1110xxxx 10xxxxxx 10xxxxxx
+            (mbc & 0xf0c0c0) == 0xe08080 &&
+            // but not 11100000 100xxxxx 10xxxxxx (overlong)
+            (mbc & 0xffe0c0) != 0xe08080 &&
+            // neither 11101101 101xxxxx 10xxxxxx (reserved for surrogates)
+            (mbc & 0xffe0c0) != 0xeda080) {
+            i+=3; continue;
+        }
+        if (s[i+2] == 0)
+            return -1;
+        // it's ok to read s[i+3] since we know s[i+2] != 0
+        mbc = mbc<<8 | s[i+3];
+        if (// 4-char: 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
+            (mbc & 0xf8c0c0c0) == 0xf0808080 &&
+            // but not 11110000 1000xxxx 10xxxxxx 10xxxxxx (overlong)
+            (mbc & 0xfff0c0c0) != 0xf0808080 &&
+            // neither 11110PPP 10PPxxxx 10xxxxxx 10xxxxxx, PPPPP>0x10 (too big)
+            (mbc & 0x07300000) <= 0x04000000) {
+            i+=4; continue;
+        }
+        return -1;
+    }
+    return i;
+}
+
+static void
+print_json_str (const char* key, const char* value) {
+    int value_len = nonfunny_utf8_len (value);
+    int key_len = strlen(key);
+    if (value_len >= 0) {
+        putchar('"');
+        fwrite (key, 1, key_len, stdout);
+        putchar('"');
+        putchar(':');
+        putchar('"');
+        fwrite (value, 1, value_len, stdout);
+        putchar ('"');
+    } else {
+        putchar('"');
+        fwrite (key, 1, key_len, stdout);
+        fwrite ("_raw\":[", 1, 7, stdout);
+        for (int i = 0; value[i] != 0; i++)
+            printf (i ? ",%d" : "%d", (unsigned int)(unsigned char)(value[i]));
+        putchar (']');
+    }
+}
+
+/* given an fd to /proc/PID and a buffer of size TASK_COMM_LEN, try to read the
+   process name. Return true on success. */
+static bool
+get_procname (int proc_fd, pid_t pid, char *procname, size_t procname_size) {
+    int fd = openat (proc_fd, "comm", O_RDONLY);
+    if (fd < 0) {
+        warn ("failed to open /proc/%u/comm", pid);
+        return false;
+    }
+    ssize_t len = read (fd, procname, procname_size - 1);
+    close (fd);
+    if (len < 0) {
+        warn ("failed to read /proc/%u/comm", pid);
+        return false;
+    }
+    while (len > 0 && procname[len-1] == '\n')
+        len--;
+    procname[len] = '\0';
+    return true;
+}
+
+/* given an fd to /proc/PID, return the parent PID if it can be determined,
+   otherwise 0. */
+static pid_t
+get_ppid (int proc_fd, pid_t pid) {
+    static char statbuf[4096];
+    int stat_fd = openat (proc_fd, "stat", O_RDONLY);
+    if (stat_fd < 0) {
+        warn ("failed to open /proc/%u/stat", pid);
+        return 0;
+    }
+    ssize_t len = read (stat_fd, statbuf, sizeof (statbuf));
+    close (stat_fd);
+    if (len < 0) {
+        warn ("failed to read /proc/%u/stat", pid);
+        return 0;
+    }
+    /* buffer is static, and read() does not nul terminate */
+    statbuf[MIN(len, sizeof (statbuf) - 1)] = '\0';
+
+    pid_t ret;
+    if (sscanf(statbuf, "%*d (%*[^)]) %*c %d", &ret) == 1)
+        return ret;
+    /* this *really* should not happen, kernel API change  */
+    warnx ("failed to parse /proc/%u/stat, please file a bug: %s", pid, statbuf);
+    return 0;
+}
+
 /**
  * print_event:
  *
@@ -240,7 +365,10 @@ print_event (const struct fanotify_event_metadata *data,
     static int procname_pid = -1;
     static char pathname[PATH_MAX];
     bool got_procname = false;
+    static char exepath[PATH_MAX];
+    bool got_exepath = false;
     struct stat proc_fd_stat = { .st_uid = -1 };
+    int ppid = 0;
 
     if ((data->mask & option_filter_mask) == 0 || !show_pid (data->pid)) {
         if (event_fd >= 0)
@@ -251,31 +379,36 @@ print_event (const struct fanotify_event_metadata *data,
     snprintf (printbuf, sizeof (printbuf), "/proc/%i", data->pid);
     int proc_fd = open (printbuf, O_RDONLY | O_DIRECTORY);
     if (proc_fd >= 0) {
+        /* get ppid */
+        if (option_parents)
+            ppid = get_ppid (proc_fd, data->pid);
+
         /* read process name */
-        int procname_fd = openat (proc_fd, "comm", O_RDONLY);
-        ssize_t len = read (procname_fd, procname, sizeof (procname));
-        if (len >= 0) {
-            while (len > 0 && procname[len-1] == '\n')
-                len--;
-            procname[len] = '\0';
+        if (get_procname (proc_fd, data->pid, procname, sizeof (procname))) {
             procname_pid = data->pid;
             got_procname = true;
-        } else {
-            debug ("failed to read /proc/%i/comm", data->pid);
         }
 
-        close (procname_fd);
-
         /* get user and group */
         if (option_user) {
             if (fstat (proc_fd, &proc_fd_stat) < 0)
-                debug ("failed to stat /proc/%i: %m", data->pid);
+                warn ("failed to stat /proc/%i", data->pid);
+        }
+
+        /* get exe */
+        if (option_exe) {
+            ssize_t len = readlinkat (proc_fd, "exe", exepath, sizeof (exepath));
+            if (len >= 0) {
+                exepath[len] = '\0';
+                got_exepath = true;
+            } else {
+                warn ("failed to readlink /proc/%i/exe", data->pid);
+            }
         }
 
         close (proc_fd);
     } else {
-        debug ("failed to open /proc/%i: %m", data->pid);
-        procname[0] = '\0';
+        warn ("failed to open /proc/%u", data->pid);
     }
 
     /* /proc/pid/comm often goes away before processing the event; reuse previously cached value if pid still matches */
@@ -289,7 +422,8 @@ print_event (const struct fanotify_event_metadata *data,
         }
     }
 
-    if (option_comm && strcmp (option_comm, procname) != 0) {
+    if (option_comm && strcmp (option_comm, procname) != 0 &&
+        procname[0] != '\0') {
         if (event_fd >= 0)
             close (event_fd);
         return;
@@ -300,18 +434,27 @@ print_event (const struct fanotify_event_metadata *data,
         event_fd = get_fid_event_fd (data);
 #endif
 
+    bool got_path = false;
+    struct stat st = { .st_uid = -1 };
     if (event_fd >= 0) {
+        if (option_json && fstat (event_fd, &st) < 0)
+            warn ("stat");
         /* try to figure out the path name */
         snprintf (printbuf, sizeof (printbuf), "/proc/self/fd/%i", event_fd);
         ssize_t len = readlink (printbuf, pathname, sizeof (pathname));
-        if (len < 0) {
-            /* fall back to the device/inode */
-            struct stat st;
-            if (fstat (event_fd, &st) < 0)
-                err (EXIT_FAILURE, "stat");
-            snprintf (pathname, sizeof (pathname), "device %i:%i inode %ld\n", major (st.st_dev), minor (st.st_dev), st.st_ino);
-        } else {
+        if (len >= 0) {
             pathname[len] = '\0';
+            got_path = true;
+        } else if (option_json) {
+          // no fallback, device/inode will always be printed
+        } else {
+            /* fall back to the device/inode */
+            if (fstat (event_fd, &st) < 0) {
+                warn ("stat");
+                pathname[0] = '\0';
+            } else {
+                snprintf (pathname, sizeof (pathname), "device %i:%i inode %ld\n", major (st.st_dev), minor (st.st_dev), st.st_ino);
+            }
         }
 
         close (event_fd);
@@ -319,21 +462,92 @@ print_event (const struct fanotify_event_metadata *data,
         snprintf (pathname, sizeof (pathname), "(deleted)");
     }
 
+    if (option_json)
+        putchar('{');
+
     /* print event */
     if (option_timestamp == 1) {
         strftime (printbuf, sizeof (printbuf), "%H:%M:%S", localtime (&event_time->tv_sec));
-        printf ("%s.%06li ", printbuf, event_time->tv_usec);
+        printf (option_json ? "\"timestamp\":\"%s.%06li\"," : "%s.%06li ", printbuf, event_time->tv_usec);
     } else if (option_timestamp == 2) {
-        printf ("%li.%06li ", event_time->tv_sec, event_time->tv_usec);
+        printf (option_json ? "\"timestamp\":%li.%06li," : "%li.%06li ", event_time->tv_sec, event_time->tv_usec);
     }
 
     /* print user and group */
     if (option_user && proc_fd_stat.st_uid != (uid_t)-1)
-        snprintf(printbuf, sizeof printbuf, " [%u:%u]", proc_fd_stat.st_uid, proc_fd_stat.st_gid);
+        snprintf(printbuf, sizeof printbuf,
+                 option_json ? "\"uid\":%u,\"gid\":%u," : " [%u:%u]",
+                 proc_fd_stat.st_uid, proc_fd_stat.st_gid);
     else
         printbuf[0] = '\0';
 
-    printf ("%s(%i)%s: %-3s %s\n", procname[0] == '\0' ? "unknown" : procname, data->pid, printbuf, mask2str (data->mask), pathname);
+    if (option_json) {
+        if (procname_pid >= 0) {
+            print_json_str("comm", procname);
+            putchar(',');
+        }
+        printf ("\"pid\":%i,%s\"types\":\"%s\"",
+                data->pid, printbuf, mask2str (data->mask));
+        if (st.st_uid != (uid_t)-1)
+            printf(",\"device\":{\"major\":%i,\"minor\":%i},\"inode\":%ld"
+                   , major (st.st_dev), minor (st.st_dev), st.st_ino);
+        if (got_path) {
+            putchar(',');
+            print_json_str("path", pathname);
+        }
+        if (option_exe && got_exepath) {
+            putchar(',');
+            print_json_str("exe", exepath);
+        }
+    } else {
+        printf ("%s(%i)%s: %-3s %s", procname[0] == '\0' ? "unknown" : procname, data->pid, printbuf, mask2str (data->mask), pathname);
+        if (option_exe && got_exepath)
+            printf (" exe=%s", exepath);
+    }
+    if (option_parents && ppid) {
+        printf(option_json ? ",\"parents\":" : ", parents");
+        char sep = option_json ? '[' : '=';
+        do {
+            printf(option_json ? "%c{\"pid\":%i" : "%c(pid=%i", sep, ppid);
+            sep = ',';
+            snprintf (printbuf, sizeof (printbuf), "/proc/%i", ppid);
+            int ppid_dir_fd = open (printbuf, O_RDONLY | O_DIRECTORY);
+            if (ppid_dir_fd >= 0) {
+                char p_procname[TASK_COMM_LEN];
+                if (get_procname (ppid_dir_fd, ppid, p_procname, sizeof (p_procname))) {
+                    if (option_json) {
+                        putchar(',');
+                        print_json_str("comm", p_procname);
+                    } else
+                      printf(" comm=%s", p_procname);
+                }
+                if (option_exe) {
+                    ssize_t len = readlinkat (ppid_dir_fd, "exe", exepath, sizeof (exepath) - 1);
+                    if (len >= 0) {
+                        exepath[len] = '\0';
+                        if (option_json) {
+                            putchar(',');
+                            print_json_str("exe", exepath);
+                        } else
+                          printf(" exe=%s", exepath);
+                    }
+                }
+                /* get next parent */
+                if (ppid == 1)
+                    ppid = 0;
+                else
+                    ppid = get_ppid (ppid_dir_fd, ppid);
+                close (ppid_dir_fd);
+            } else {
+                warn ("failed to open %s", printbuf);
+                ppid = 0;
+            }
+            putchar(option_json ? '}' : ')');
+        } while (ppid > 0);
+        if (option_json)
+            putchar(']');
+    }
+    printf(option_json ? "}\n" : "\n");
 }
 
 static void
@@ -439,8 +653,11 @@ help (void)
 "  -t, --timestamp\t\tAdd timestamp to events. Give twice for seconds since the epoch.\n"
 "  -u, --user\t\t\tAdd user ID and group ID to events.\n"
 "  -p PID, --ignore-pid PID\tIgnore events for this process ID. Can be specified multiple times.\n"
-"  -f TYPES, --filter=TYPES\tShow only the given event types; choose from C, R, O, or W, e. g. --filter=OC.\n"
+"  -f TYPES, --filter=TYPES\tShow only the given event types; choose from C, R, O, W, +, D, < or >, e. g. --filter=OC.\n"
 "  -C COMM, --command=COMM\tShow only events for this command.\n"
+"  -j, --json\t\t\tWrite events in JSONL format.\n"
+"  -P, --parents\t\t\tInclude information about all parent processes.\n"
+"  -e, --exe\t\t\tAdd executable path to events.\n"
 "  -h, --help\t\t\tShow help.");
 }
 
@@ -466,12 +683,15 @@ parse_args (int argc, char** argv)
         {"ignore-pid",    required_argument, 0, 'p'},
         {"filter",        required_argument, 0, 'f'},
         {"command",       required_argument, 0, 'C'},
+        {"json",          no_argument,       0, 'j'},
+        {"parents",       no_argument,       0, 'P'},
+        {"exe",           no_argument,       0, 'e'},
         {"help",          no_argument,       0, 'h'},
         {0,               0,                 0,  0 }
     };
 
     while (1) {
-        c = getopt_long (argc, argv, "C:co:s:tup:f:h", long_options, NULL);
+        c = getopt_long (argc, argv, "C:co:s:tup:f:jPeh", long_options, NULL);
 
         if (c == -1)
             break;
@@ -479,6 +699,8 @@ parse_args (int argc, char** argv)
         switch (c) {
             case 'C':
                 option_comm = strdup (optarg);
+                if (!option_comm)
+                    err(EXIT_FAILURE, "memory allocation failed for --command");
                 /* see https://man7.org/linux/man-pages/man5/proc_pid_comm.5.html */
                 if (strlen (option_comm) > TASK_COMM_LEN - 1) {
                     option_comm[TASK_COMM_LEN - 1] = '\0';
@@ -487,15 +709,17 @@ parse_args (int argc, char** argv)
                 break;
 
             case 'c':
-                option_current_mount = 1;
+                option_current_mount = true;
                 break;
 
             case 'o':
                 option_output = strdup (optarg);
+                if (!option_output)
+                    err(EXIT_FAILURE, "memory allocation failed for --output");
                 break;
 
             case 'u':
-                option_user = 1;
+                option_user = true;
                 break;
 
             case 'f':
@@ -546,7 +770,8 @@ parse_args (int argc, char** argv)
                 pid = strtol (optarg, &endptr, 10);
                 if (*endptr != '\0' || pid <= 0)
                     errx (EXIT_FAILURE, "Error: Invalid PID");
-                if (ignored_pids_len < sizeof (ignored_pids))
+                if (ignored_pids_len
+                    < (sizeof (ignored_pids) / sizeof (ignored_pids[0])))
                     ignored_pids[ignored_pids_len++] = pid;
                 else
                     errx (EXIT_FAILURE, "Error: Too many ignored PIDs");
@@ -557,6 +782,18 @@ parse_args (int argc, char** argv)
                     errx (EXIT_FAILURE, "Error: --timestamp option can be given at most two times");
                 break;
 
+            case 'j':
+                option_json = true;
+                break;
+
+            case 'P':
+                option_parents = true;
+                break;
+
+            case 'e':
+                option_exe = true;
+                break;
+
             case 'h':
                 help ();
                 exit (EXIT_SUCCESS);
